@@ -3,6 +3,7 @@ from __future__ import print_function
 from collections import namedtuple, defaultdict
 import copy
 import json
+from os.path import basename
 import struct
 
 import numpy as np
@@ -13,7 +14,7 @@ from ... import compression
 from ...lib import jsonify, toiter
 from ...lru import LRU
 from ...exceptions import SpecViolation, EmptyFileException
-from ...storage import SimpleStorage, Storage
+from ...storage import SimpleStorage, GreenStorage, Storage
 
 ShardLocation = namedtuple('ShardLocation', 
   ('shard_number', 'minishard_number', 'remainder')
@@ -217,6 +218,7 @@ class ShardReader(object):
     self, meta, cache, spec,
     shard_index_cache_size=512,
     minishard_index_cache_size=128,
+    green=False
   ):
     """
     Reads standard Precomputed shard files. 
@@ -231,6 +233,7 @@ class ShardReader(object):
     self.meta = meta
     self.cache = cache
     self.spec = spec
+    self.green = green
 
     self.shard_index_cache = LRU(shard_index_cache_size)
     self.minishard_index_cache = LRU(minishard_index_cache_size)
@@ -255,45 +258,168 @@ class ShardReader(object):
     Returns: 2^minishard_bits entries of a uint64 
       array of [[ byte start, byte end ], ... ] 
     """
-    index_path = self.meta.join(path, filename)
-    alias_path = self.meta.join(path, filename.replace('.shard', '.index'))
+    indices = self.get_indices([ filename ], path, progress=False)
+    index = list(indices.values())[0]
+    if index is None:
+      raise EmptyFileException(filename + " was zero bytes.")
+    return index
 
-    if filename in self.shard_index_cache:
-      return self.shard_index_cache[filename]
+  def get_indices(self, filenames, path="", progress=None):
+    """
+    For all given files, retrieves the shard index which 
+    is used for locating the appropriate minishard indices.
 
-    index_length = self.spec.index_length()
+    Returns: { 
+      path_to_/filename.shard: 2^minishard_bits entries of a uint64 
+            array of [[ byte start, byte end ], ... ],
+      ...
+    } 
+    """
+    filenames = toiter(filenames)
+    filenames = [ self.meta.join(path, fname) for fname in filenames ]
+    fufilled = { 
+      fname: self.shard_index_cache[fname] \
+      for fname in filenames \
+      if fname in self.shard_index_cache  
+    }
 
-    binary = self.cache.download_single_as(
-      index_path, alias_path,
-      start=0, end=index_length,
-      compress=False
-    )
+    requests = []
+    for fname in filenames:
+      if fname in fufilled:
+        continue
+      requests.append({
+        'path': fname, 
+        'local_alias': fname + '.index', 
+        'start': 0, 
+        'end': self.spec.index_length(),
+      })
 
+    progress = 'Shard Indices' if progress else False
+    binaries = self.cache.download_as(requests, progress=progress)
+    for (fname, start, end), content in binaries.items():
+      try:
+        index = self.decode_index(content, fname)
+        self.shard_index_cache[fname] = index
+        fufilled[fname] = index
+      except EmptyFileException:
+        self.shard_index_cache[fname] = None
+        fufilled[fname] = None
+
+    return fufilled
+
+  def decode_index(self, binary, filename='Shard'):
     if binary is None or len(binary) == 0:
       raise EmptyFileException(filename + " was zero bytes.")
-    elif len(binary) != index_length:
+    elif len(binary) != self.spec.index_length():
       raise SpecViolation(
-        filename + " was an incorrect length ({}) for this specification ({}).".format(
-          len(binary), index_length
+        filename + ": shard index was an incorrect length ({}) for this specification ({}).".format(
+          len(binary), self.spec.index_length()
         ))
     
     index = np.frombuffer(binary, dtype=np.uint64)
     index = index.reshape( (index.size // 2, 2), order='C' )
-    self.shard_index_cache[filename] = index
-    return index
+    return index + self.spec.index_length()
+
+  def decode_minishard_index(self, minishard_index, filename=''):
+    """Returns [[label, offset, size], ... ] where offset and size are in bytes."""
+
+    if self.spec.minishard_index_encoding != 'raw':
+      minishard_index = compression.decompress(
+        minishard_index, encoding=self.spec.minishard_index_encoding, filename=filename
+      )
+
+    minishard_index = np.copy(np.frombuffer(minishard_index, dtype=np.uint64))
+    minishard_index = minishard_index.reshape( (3, len(minishard_index) // 3), order='C' ).T
+
+    for i in range(1, minishard_index.shape[0]):
+      minishard_index[i, 0] += minishard_index[i-1, 0]
+      minishard_index[i, 1] += minishard_index[i-1, 1] + minishard_index[i-1, 2]
+
+    minishard_index[:,1] += self.spec.index_length()
+
+    return minishard_index 
 
   def get_minishard_index(self, filename, index, minishard_no, path=""):
-    res = self.get_minishard_indices(filename, index, minishard_no, path)
-    return res[minishard_no]
-
-  def get_minishard_indices(self, filename, index, minishard_nos, path=""):
     """
     Retrieves the minishard index for a given minishard number.
 
     Returns: uint64 Nx3 array with multiple rows of [segid, byte start, byte end]
     """
+    res = self.get_minishard_indices(filename, index, minishard_no, path)
+    return res[minishard_no]
+
+  def get_minishard_indices(self, filename, index, minishard_nos, path=""):
+    """
+    Retrieves the minishard indices for a set of minishard numbers.
+
+    Returns: { minishard_no: uint64 Nx3 array of [segid, byte start, byte end], ... }
+    """
+    res = self.get_minishard_indices_for_files(( (filename, index, minishard_nos), ), path)
+    return res[filename]
+
+  def get_minishard_indices_for_files(self, requests, path="", progress=None):
+    """
+    Fetches the specified minishard indices for all the specified files
+    at once. This is required to get high performance as opposed to fetching
+    the all minishard indices for a single file.
+
+    requests: iterable of tuples
+      [  (filename, index, minishard_numbers), ... ]
+
+    Returns: map of filename -> minishard numbers -> minishard indices
+
+    e.g. 
+    {
+      filename_1: {
+          0: uint64 Nx3 array of [segid, byte start, byte end],
+          1: ...,
+      }
+      filename_2: ...
+    }
+    """
+    fufilled_by_filename = defaultdict(dict)
+    msn_map = {}
+
+    download_requests = []
+    for filename, index, minishard_nos in requests:
+      fufilled_requests, pending_requests = self.compute_minishard_index_requests(
+        filename, index, minishard_nos, path
+      ) 
+      fufilled_by_filename[filename] = fufilled_requests
+      for msn, start, end in pending_requests:
+        msn_map[(filename, start, end)] = msn
+
+        filepath = self.meta.join(path, filename)
+
+        download_requests.append({
+          'path': filepath,
+          'local_alias': '{}-{}.msi'.format(filepath, msn),
+          'start': start,
+          'end': end,
+        })
+
+    progress = 'Minishard Indices' if progress else False
+    results = self.cache.download_as(download_requests, progress=progress)
+  
+    for (filename, start, end), content in results.items():
+      filename = basename(filename)
+      cache_key = (filename, start, end)
+      msn = msn_map[cache_key]
+      minishard_index = self.decode_minishard_index(content, filename)
+      self.minishard_index_cache[cache_key] = minishard_index
+      fufilled_by_filename[filename][msn] = minishard_index
+
+    return fufilled_by_filename
+
+  def compute_minishard_index_requests(self, filename, index, minishard_nos, path=""):
+    """
+    Helper method for get_minishard_indices_for_files. 
+    Computes which requests must be made over the network vs can be fufilled from LRU cache.
+    """
     minishard_nos = toiter(minishard_nos)
-    index_offset = self.spec.index_length()
+
+    if index is None:
+      return ({ msn: None for msn in minishard_nos }, [])
 
     fufilled_requests = {}
 
@@ -306,8 +432,6 @@ class ShardReader(object):
         fufilled_requests[msn] = None
         continue
 
-      bytes_start += index_offset
-      bytes_end += index_offset
       bytes_start, bytes_end = int(bytes_start), int(bytes_end)
       byte_ranges[msn] = (bytes_start, bytes_end)
 
@@ -319,45 +443,11 @@ class ShardReader(object):
       if cache_key in self.minishard_index_cache:
         fufilled_requests[msn] = self.minishard_index_cache[cache_key]
       else:
-        pending_requests.append((bytes_start, bytes_end))
+        pending_requests.append((msn, bytes_start, bytes_end))
 
-    StorageClass = SimpleStorage if len(pending_requests) == 1 else Storage
+    return (fufilled_requests, pending_requests)
 
-    with StorageClass(full_path) as stor:
-      filenames = ( filename for _ in pending_requests )
-      starts = ( start for (start, end) in pending_requests )
-      ends = ( end for (start, end) in pending_requests )
-      results = stor.get_files(filenames, starts, ends)
-  
-    del pending_requests
-
-    def decode_minishard_index(minishard_index):
-      if self.spec.minishard_index_encoding != 'raw':
-        minishard_index = compression.decompress(
-          minishard_index, encoding=self.spec.minishard_index_encoding, filename=filename
-        )
-
-      minishard_index = np.copy(np.frombuffer(minishard_index, dtype=np.uint64))
-      minishard_index = minishard_index.reshape( (3, len(minishard_index) // 3), order='C' ).T
-
-      for i in range(1, minishard_index.shape[0]):
-        minishard_index[i, 0] += minishard_index[i-1, 0]
-        minishard_index[i, 1] += minishard_index[i-1, 1] + minishard_index[i-1, 2]
-
-      return minishard_index 
-
-    byte_ranges = { v: k for k, v in byte_ranges.items() }
-    for res in results:
-      start, end = res['byte_range']
-      msn = byte_ranges[(start, end)]
-      cache_key = (filename, start, end)
-      minishard_index = decode_minishard_index(res['content'])
-      self.minishard_index_cache[cache_key] = minishard_index
-      fufilled_requests[msn] = minishard_index
-
-    return fufilled_requests
-
-  def exists(self, labels, path="", return_byte_range=False):
+  def exists(self, labels, path="", return_byte_range=False, progress=None):
     """
     Checks a shard's minishard index for whether a file exists.
 
@@ -379,91 +469,166 @@ class ShardReader(object):
     except TypeError:
       return_one = True
 
-    results = {}
+    to_labels = defaultdict(list)
+    to_all_labels = defaultdict(list)
+    filename_to_minishard_num = defaultdict(list)
+
     for label in set(toiter(labels)):
       filename, minishard_number = self.compute_shard_location(label)
-      
+      to_labels[(filename, minishard_number)].append(label)
+      to_all_labels[filename].append(label)
+      filename_to_minishard_num[filename].append(minishard_number)
+
+    indices = self.get_indices(to_all_labels.keys(), path, progress=progress)
+
+    all_minishards = self.get_minishard_indices_for_files([ 
+      (basename(filepath), index, filename_to_minishard_num[basename(filepath)]) \
+      for filepath, index in indices.items()
+    ], path, progress=progress)
+
+    results = {}
+    for filename, file_minishards in all_minishards.items():
       filepath = self.meta.join(path, filename)
+      for mini_no, msi in file_minishards.items():
+        labels = to_labels[(filename, mini_no)]
 
-      if self.cache.enabled:
-        cached = self.cache.has(self.meta.join(path, str(label)), progress=False)
-        if cached is not None:
-          results[label] = filepath
-          continue
+        for label in labels:
+          if msi is None:
+            results[label] = None
+            continue
 
-      index = self.get_index(filename, path)
-
-      minishard_index = self.get_minishard_index(
-        filename, index, 
-        minishard_number, path
-      )
-
-      if minishard_index is None:
-        results[label] = None
-        continue
-
-      idx = np.where(minishard_index[:,0] == label)[0]
-      if len(idx) == 0:
-        results[label] = None
-      else:
-        if return_byte_range:
-          _, offset, size = minishard_index[idx,:][0]
-          index_offset = self.spec.index_length()
-          results[label] = [ filepath, int(offset + index_offset), int(size) ]
-        else:
-          results[label] = filepath
+          idx = np.where(msi[:,0] == label)[0]
+          if len(idx) == 0:
+            results[label] = None
+          else:
+            if return_byte_range:
+              _, offset, size = msi[idx,:][0]
+              results[label] = [ filepath, int(offset), int(size) ]
+            else:
+              results[label] = filepath
 
     if return_one:
-      return results[label]
+      return next(list(results.values()))
     return results
 
-  def get_data(self, label, path=""):
-    filename, minishard_number = self.compute_shard_location(label)
-    
+  def disassemble_shard(self, shard):
+    """
+    Given an entire shard as a bytestring, convert 
+    it into a dict of { label: byte content }.
+    """
+    index = self.decode_index(shard[:self.spec.index_length()])
+    shattered = {}
+    for start, end in index:
+      start, end = int(start), int(end)
+      if start == end:
+        continue
+
+      msi = self.decode_minishard_index(shard[start:end])
+      for label, offset, size in msi:
+        offset, size = int(offset), int(size)
+        binary = shard[offset:offset+size]
+        
+        if self.spec.data_encoding != 'raw':
+          binary = compression.decompress(binary, encoding=self.spec.data_encoding)
+        
+        shattered[label] = binary
+
+    return shattered
+
+  def get_data(self, label, path="", progress=None):
+    """Fetches data from shards.
+
+    label: one or more segment ids
+    path: subdirectory path
+    progress: display progress bars
+
+    Return: 
+      if label is a scalar:
+        a byte string
+      else: (label is an iterable)
+        {
+          label_1: byte string,
+          ....
+        }
+    """
+    label, return_multiple = toiter(label, is_iter=True)
+    label = set(( int(l) for l in label))
+
+    results = {}
     if self.cache.enabled:
-      cached = self.cache.get_single(self.meta.join(path, str(label)), progress=False)
-      if cached is not None:
-        return cached
+      results = self.cache.get([ 
+        self.meta.join(path, str(lbl)) for lbl in label
+      ], progress=progress)
 
-    index = self.get_index(filename, path)
+    for cloudpath, content in results.items():
+      if content is None:
+        label.remove(int(basename(cloudpath)))
 
-    minishard_index = self.get_minishard_index(
-      filename, index, 
-      minishard_number, path
-    )
 
-    if minishard_index is None:
-      return None
+    # { label: [ filename, byte start, num_bytes ] }
+    exists = self.exists(label, path, return_byte_range=True, progress=progress)
+    for k in list(exists.keys()):
+      if exists[k] is None:
+        results[k] = None
+        del exists[k]
 
-    idx = np.where(minishard_index[:,0] == label)[0]
-    if len(idx) == 0:
-      return None
-    else:
-      idx = idx[0]
+    key_label = { (basename(v[0]), v[1], v[2]): k for k,v in exists.items() }
 
-    _, offset, size = minishard_index[idx,:]
+    filenames = ( basename(ext[0]) for ext in exists.values() )
+    starts = ( int(ext[1]) for ext in exists.values() )
+    ends = ( int(ext[1]) + int(ext[2]) for ext in exists.values() )
 
-    index_offset = self.spec.index_length()
-    offset = int(offset + index_offset)
-       
     full_path = self.meta.join(self.meta.cloudpath, path)
 
-    with SimpleStorage(full_path) as stor:
-      binary = stor.get_file(filename, start=offset, end=int(offset + size))
+    StorageClass = Storage
+    if self.green:
+      StorageClass = GreenStorage
+    elif len(exists) == 1:
+      StorageClass = SimpleStorage
+
+    with StorageClass(full_path, progress=progress) as stor:
+      remote_files = stor.get_files(filenames, starts=starts, ends=ends)
+
+    binaries = {}
+    for res in remote_files:
+      if res['error']:
+        raise res['error']
+      start, end = res['byte_range']
+      key = (res['filename'], start, end - start)
+      lbl = key_label[key]
+      binaries[lbl] = res['content']
+    del remote_files
 
     if self.spec.data_encoding != 'raw':
-      binary = compression.decompress(binary, encoding=self.spec.data_encoding, filename=filename)
+      for filepath, binary in tqdm(binaries.items(), desc="Decompressing", disable=(not progress)):
+        if binary is None:
+          continue
+        binaries[filepath] = compression.decompress(
+          binary, encoding=self.spec.data_encoding, filename=filepath
+        )
       
     if self.cache.enabled:
-      self.cache.put_single(self.meta.join(path, str(label)), binary, progress=False)
+      self.cache.put([ 
+        (filepath, binary) for filepath, binary in binaries.items()
+      ], progress=progress)
 
-    return binary
+    results.update(binaries)
 
-  def list_labels(self, filename, path=""):
+    if return_multiple:
+      return results
+    return next(iter(results.values()))
+
+  def list_labels(self, filename, path="", size=False):
     """
     List all the labels in the index of a given shard file.
 
-    Returns: np.uint64 array
+    size: (bool) if True, list the size in bytes of each label
+
+    Returns: 
+      if not size:
+        np.uint64 array of labels 
+      else:
+        [ (label, size in bytes), ... ] in descending order of size
     """
     index = self.get_index(filename, path)
     all_minishard_nos = list(range(len(index)))
@@ -471,10 +636,18 @@ class ShardReader(object):
     minishard_indicies = [  
       msi for msi in minishard_indicies.values() if msi is not None
     ]
-    labels = np.concatenate([  
-      msi[:,0] for msi in minishard_indicies
-    ])
-    return np.sort(labels)
+    if not size:
+      labels = np.concatenate([  
+        msi[:,0] for msi in minishard_indicies
+      ])
+      return np.sort(labels)
+    else:
+      labels = np.concatenate([  
+        msi[:,:]
+        for msi in minishard_indicies
+      ])
+      labels = [ (row[0], row[2]) for row in labels[:] ]
+      return sorted(labels, key=lambda x: x[1], reverse=True)
 
 def synthesize_shard_files(spec, data, progress=False):
   """
